/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_env_variable.h"

#include "initng_also.h"

s_entry ALSO_START = { "also_start", STRINGS, NULL,
    "When this service is starting, also start this."
};
s_entry ALSO_STOP = { "also_stop", STRINGS, NULL,
    "When this service is stopping, also stop this."
};

int service_state(active_db_h * service)
{
    const char *tmp = NULL;
    active_db_h *current = NULL;

    assert(service);
    assert(service->name);

    /* if service is loading, start all in ALSO_START */
    if (IS_MARK(service, &LOADING))
    {
        tmp = NULL;
        while ((tmp =
                initng_active_db_get_next_string(&ALSO_START, service, tmp)))
        {
            char *fixed = NULL;

            fixed = fix_variables(tmp, service);
            if (initng_active_db_find_by_name(fixed))
            {
                free(fixed);
                continue;
            }

            /* HACK WARNING
             * service name starting with "virtual/" is no idea to start, sience
             * they wont exists anyway, this is a waist of resources.
             */
            if (strncmp(fixed, "virtual/", 8) == 0)
            {
                free(fixed);
                continue;
            }

            if (!initng_handler_start_new_service_named(fixed))
            {
                F_("%s also_start %s could not start!\n", service->name,
                   fixed);
                initng_handler_stop_service(service);
                free(fixed);
                return (FALSE);
            }
        }
        return (TRUE);
    }

    /* if this service is stopping, stopp all in ALSO_STOP */
    if (IS_STOPPING(service))
    {
        /* Handle ALSO_STOP */
        tmp = NULL;
        while ((tmp =
                initng_active_db_get_next_string(&ALSO_STOP, service, tmp)))
        {
            char *fixed = NULL;

            fixed = fix_variables(tmp, service);
            if ((current = initng_active_db_find_by_name(fixed)))
            {
                if (g.sys_state != STATE_STOPPING)
                {
                    W_("service %s also stops %s\n", service->name, fixed);
                }
                initng_handler_stop_service(current);
            }
            free(fixed);
        }
        return (TRUE);
    }

    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&ALSO_START);
    initng_service_data_types_add(&ALSO_STOP);
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 50, &service_state);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_service_data_types_del(&ALSO_START);
    initng_service_data_types_del(&ALSO_STOP);
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &service_state);
}
