/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>
#include <time.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_execute.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_static_process_types.h"

#include "initng_cron.h"


static int start_CRON(active_db_h * service_to_start);
static int stop_CRON(active_db_h * service_to_stop);




static void killed_cron(active_db_h * service, process_h * process);

stype_h TYPE_CRON = { "cron", start_CRON, stop_CRON, NULL };
ptype_h T_CRON = { "cron", &killed_cron };

/*s_entry FIXED = { "fixed", STRING, CRON_OPT }; */
s_entry MONTHLY = { "monthly", SET, &TYPE_CRON, "Run this cron every month."
};
s_entry WEEKLY = { "weekly", SET, &TYPE_CRON, "Run this cron every week." };
s_entry DAILY = { "daily", SET, &TYPE_CRON, "Run this cron every day." };
s_entry HOURLY = { "hourly", SET, &TYPE_CRON, "Run this cron every hour." };
s_entry LAST_RUN = { NULL, INT, NULL, NULL };


static int start_CRON(active_db_h * service_to_start)
{
    if (!initng_common_mark_service(service_to_start, &WAIT_FOR_CRON))
    {
        W_("mark_service WAIT_FOR_CRON failed for cron-service %s\n",
           service_to_start->name);
        return (FALSE);
    }
    return (TRUE);
}

static int stop_CRON(active_db_h * service_to_stop)
{
    /* cron status is handled specially */
    if (IS_WAITING(service_to_stop))
    {
        initng_common_mark_service(service_to_stop, &STOPPED);
        return (TRUE);
    }
    return (FALSE);
}

static void run_cron(active_db_h * service)
{
    if (!initng_common_mark_service(service, &RUNNING))
        return;

    initng_execute_launch(service, &T_CRON);
}

static void got_alarm(void)
{
    active_db_h *current = NULL;

    while_active_db(current)
    {
        time_t last;
        time_t now;

        if (current->type != &TYPE_CRON)
            continue;
        now = time(0);
        last = (time_t) initng_active_db_get_int(&LAST_RUN, current);

        if (initng_active_db_is(&HOURLY, current))
        {
            /* 60 seconds one minute */
            /* 60 minutes, one hour */
            if (now - last > (60 * 60))
                run_cron(current);
            else
                alarm(now - last);
            continue;
        }

        if (initng_active_db_is(&DAILY, current))
        {
            /* 60 seconds one minute */
            /* 60 minutes, one hour */
            if (now - last > (60 * 60 * 24))
                run_cron(current);
            else
                alarm(now - last);
            continue;
        }

        if (initng_active_db_is(&WEEKLY, current))
        {
            /* 60 seconds one minute */
            /* 60 minutes, one hour */
            if (now - last > (60 * 60 * 24 * 7))
                run_cron(current);
            else
                alarm(now - last);
            continue;
        }

        if (initng_active_db_is(&MONTHLY, current))
        {
            /* 60 seconds one minute */
            /* 60 minutes, one hour */
            if (now - last > (60 * 60 * 24 * 7 * 30))
                run_cron(current);
            else
                alarm(now - last);
            continue;
        }

    }
}

static void killed_cron(active_db_h * service, process_h * process)
{
    if (process->r_code == 0)
        printf("Cron for service %s, exited sucessfully!\n", service->name);
    else
        printf("Cron for service %s, exited with r_code %i!\n", service->name,
               process->r_code);


    list_del(&process->list);
    initng_process_db_free(process);

    if (!initng_common_mark_service(service, &WAIT_FOR_CRON))
    {
        F_("Unable to reset cron.\n");
    }

    return;
}

static int set_alarm(active_db_h * service)
{
    if (!IS_MARK(service, &WAIT_FOR_CRON))
        return (TRUE);

    if (service->type != &TYPE_CRON)
    {
        F_("Service %s has status CRON_WAIT but is not a TYPE_CRON!\n");
        return (FALSE);
    }


    if (initng_active_db_is(&HOURLY, service))
    {
        initng_active_db_set_int(&LAST_RUN, service, time(0));
        alarm(60 * 60);
    }

    if (initng_active_db_is(&DAILY, service))
    {
        initng_active_db_set_int(&LAST_RUN, service, time(0));
        alarm(60 * 60 * 24);
    }

    if (initng_active_db_is(&WEEKLY, service))
    {
        initng_active_db_set_int(&LAST_RUN, service, time(0));
        alarm(60 * 60 * 24 * 7);
    }

    if (initng_active_db_is(&MONTHLY, service))
    {
        initng_active_db_set_int(&LAST_RUN, service, time(0));
        alarm(60 * 60 * 24 * 7 * 30);
    }

    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_types_add(&TYPE_CRON);

    /*initng_sdt_add(&FIXED); */
    initng_service_data_types_add(&MONTHLY);
    initng_service_data_types_add(&WEEKLY);
    initng_service_data_types_add(&DAILY);
    initng_service_data_types_add(&HOURLY);

    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 80, &set_alarm);
    initng_plugin_hook_add(&g.ALARM, 80, &got_alarm);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    /*initng_sdt_del(&FIXED); */
    initng_service_types_del(&TYPE_CRON);
    initng_service_data_types_del(&MONTHLY);
    initng_service_data_types_del(&WEEKLY);
    initng_service_data_types_del(&DAILY);
    initng_service_data_types_del(&HOURLY);
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &set_alarm);
    initng_plugin_hook_del(&g.ALARM, &got_alarm);
}
