
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>
#include <dlfcn.h>

#include "../../src/initng_active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_execute.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_common.h"
#include "../../src/initng_signal.h"
#include "../../src/initng_fork.h"

#include "initng_dllaunch.h"

static int initng_dl_exec(active_db_h * s, process_h * p, const char *name);

s_entry DL = { "dl", STRING, NULL,
    "Load a shared library and execute."
};

static int initng_d(active_db_h * service, process_h * process)
{
    const char *e = NULL;

    assert(service);
    assert(service->name);

    if ((e = initng_active_db_get_string(&DL, service)))
    {
        return (initng_dl_exec(service, process, e));
    }

    return (FALSE);
}



static int initng_dl_exec(active_db_h * s, process_h * p, const char *name)
{
    pid_t pid_fork;             /* pid got from fork() */

    assert(s);
    assert(s->name);
    assert(p);
    assert(name);

    if ((pid_fork = initng_fork(s, p)) == 0)
    {
        void *dlhandle;
        struct stat st;

        int (*function) (active_db_h * service);

        if (stat(name, &st) != 0 || !S_ISREG(st.st_mode))
        {
            F_("Could not find file %s\n", name);
            _exit(1);
        }

        dlerror();                          /* clear errors */
        dlhandle = dlopen(name, RTLD_LAZY);


        if (!dlhandle)
        {
            F_("Problem loading dl_start module \"%s\"\n", dlerror());
            _exit(2);
        }

        dlerror();

        function = dlsym(dlhandle, p->pt->name);
        if (!function)
        {
            F_("Could not find start or stop function!\n");
            dlclose(dlhandle);
            _exit(3);
        }

        _exit((*function) (s));
        _exit(99);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);
    p->pid = pid_fork;
    return pid_fork;

}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_plugin_hook_add(&g.LAUNCH, 50, &initng_d);

    initng_service_data_types_add(&DL);

    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");

    initng_plugin_hook_del(&g.LAUNCH, &initng_d);
    initng_service_data_types_del(&DL);

}
