
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include "../../src/initng_global.h"

#include "../../src/initng_string_tools.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_static_service_types.h"

#include "initng_d_parser.h"

static void set_d_parser(s_entry * type, char *value,
                         service_cache_h * from_service);
static void string_d_parser(s_entry * type, char *value,
                            service_cache_h * from_service);
static void strings_d_parser(s_entry * type, char *value,
                             service_cache_h * from_service);
static void int_d_parser(s_entry * type, char *value,
                         service_cache_h * from_service);


static void set_d_parser(s_entry * type, char *value
                         __attribute__ ((unused)),
                         service_cache_h * from_service)
{
    assert(from_service);
    assert(from_service->name);
    assert(value);

    D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    initng_service_cache_set(type, from_service);
}


static void string_d_parser(s_entry * type, char *value,
                            service_cache_h * from_service)
{
    int chars = 0;

    assert(from_service);
    assert(from_service->name);
    assert(value);

    while (value[chars] && value[chars] != '\n')
        chars++;

    initng_service_cache_set_string(type, from_service,
                                    i_strndup(value, chars));
}

static void int_d_parser(s_entry * type, char *value,
                         service_cache_h * from_service)
{
    int chars = 0;
    char *to = NULL;

    assert(from_service);
    assert(from_service->name);
    assert(value);

    while (value[chars] && value[chars] != '\n')
        chars++;


    initng_service_cache_set_int(type, from_service, atoi(to));

    if (to)
        free(to);
}

static void strings_d_parser(s_entry * type, char *value,
                             service_cache_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    assert(from_service->name);
    assert(value);
    D_("string_parser(%s,s,%s);\n", type->opt_name, from_service->name);

    while ((to = st_dup_next_word(&value)))
    {
        D_("Adding %s\n", to);
        inting_service_cache_set_another_string(type, from_service, to);
        JUMP_SPACES(value);
    }
}

static void read_opt(char *tmp, service_cache_h * serv)
{
    s_entry *current = NULL;
    int w_l = 0;

    assert(tmp);
    assert(serv);
    assert(serv->name);

    while (tmp[w_l] && tmp[w_l] != ' ')
        w_l++;

    /* browse option table */
    while_service_data_types(current)
    {
        s_entry *check = current;

        /* if we found it */
        if (!current->opt_name)
            continue;
        if (w_l != current->opt_name_len)
            continue;
        if (strncmp(current->opt_name, tmp, w_l) != 0)
            continue;

        /* check the option type */
        if (current->ot && current->ot != &TYPE_SERVICE)
            continue;

        D_("Found it\n");
        tmp = tmp + w_l + 1;
        /* jump forward all spaces */
        while (tmp[0] == ' ')
            tmp++;

        if (!tmp[0])
            return;

        while (check->opt_type == ALIAS && check->alias)
            check = check->alias;

        switch (check->opt_type)
        {
            case STRING:
            case VARIABLE_STRING:
                string_d_parser(current, tmp, serv);
                return;
            case STRINGS:
            case VARIABLE_STRINGS:
                strings_d_parser(current, tmp, serv);
                return;
            case SET:
                set_d_parser(current, tmp, serv);
                return;
            case INT:
                int_d_parser(current, tmp, serv);
                return;
            default:
                return;
        }
    }
}

/* Load a service from a process_name or process_path */
service_cache_h *initng_d_parser(const char *s_name)
{
    char filetoparse[200];
    service_cache_h *serv = NULL;
    char *content = NULL;
    char *tmp = NULL;

    s_entry *START = initng_service_data_types_find("start");
    s_entry *START_ARGS = initng_service_data_types_find("start_args");
    s_entry *STOP = initng_service_data_types_find("stop");
    s_entry *STOP_ARGS = initng_service_data_types_find("stop_args");

    assert(s_name);
    D_("initng_d_parser(%s);\n", s_name);

    if (strcmp(s_name, "reboot") == 0 || strcmp(s_name, "halt") == 0)
        return (NULL);


    strcpy(filetoparse, "/etc/init.d/");
    strcat(filetoparse, s_name);

    /* read config file */
    if (!open_read_close(filetoparse, &content))
    {
        D_("Can't open config file %s.\n", filetoparse);
        return (NULL);
    }
    D_("Did read the complete file!\n");

    serv =
        initng_service_cache_new(s_name, initng_service_types_get("service"));
    if (!serv)
    {
        D_("Can't allocate service\n");
        free(content);
        return (NULL);
    }
    D_("Service allocated!\n");

    initng_service_cache_set_string(&FROM_FILE, serv, i_strdup(filetoparse));
    inting_service_cache_set_another_string(START, serv,
                                            i_strdup(filetoparse));
    inting_service_cache_set_another_string(STOP, serv,
                                            i_strdup(filetoparse));
    initng_service_cache_set_string(START_ARGS, serv, i_strdup("start"));
    initng_service_cache_set_string(STOP_ARGS, serv, i_strdup("stop"));
    D_("Standard variables set, now parsing.\n");

    tmp = content;
    while ((tmp = strstr(tmp, "#% ")))
    {
        D_("Found magic word data entry.\n");
        /* jump to the first char */
        tmp = tmp + 3;
        while (tmp[0] == ' ')
            tmp++;

        read_opt(tmp, serv);

    }

    free(content);
    D_("Parse done, making sure need is set.\n");

    /* check if need flag is set */
    if (g.i_am_init == TRUE && !initng_service_cache_is(&NEED, serv))
    {
        inting_service_cache_set_another_string(&NEED, serv,
                                                i_strdup("system/initial"));
        inting_service_cache_set_another_string(&NEED, serv,
                                                i_strdup("system/mountroot"));

    }

    D_("Now adding...\n");
    /* add to service db */
    if (!initng_service_cache_add(serv))
    {
        F_("can't add service to service db!\n");
        if (serv->name)
            free(serv->name);

        initng_service_cache_remove_all(serv);

        if (serv)
            free(serv);
        return (NULL);
    }

    D_("Now added, returning it\n");
    return (serv);
}

int module_init(const char *version)
{

    D_("d_parser: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    return (initng_plugin_hook_add(&g.PARSERS, 80, &initng_d_parser));

}

void module_unload(void)
{
    D_("d_parser: module_unload();\n");
    initng_plugin_hook_del(&g.PARSERS, &initng_d_parser);
}
