
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>

#define XML_EXT ".xml"
#include "../../initng-paths.h"
#include "../../src/initng_global.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_service_types.h"

#include "initng_ezxml_parser.h"
#include "ezxml.h"

static service_cache_h *initng_parse_ezxml_file(const char *filename,
                                                const char *watch_for);

static service_cache_h *initng_parse_service_ezxml(ezxml_t node,
                                                   const char *watch_for,
                                                   service_cache_h * father,
                                                   stype_h * type);

static int parse_xml_opt(const char *name, const char *value, stype_h * stype,
                         service_cache_h * srv);


static int set_xml_parser(s_entry * type, const char *value,
                          service_cache_h * from_service);
static int string_xml_parser(s_entry * type, const char *value,
                             service_cache_h * from_service);
static int strings_xml_parser(s_entry * type, const char *value,
                              service_cache_h * from_service);
static int int_xml_parser(s_entry * type, const char *value,
                          service_cache_h * from_service);


/* Load a service from a process_name or process_path */
service_cache_h *initng_ezxml_load_process(const char *process_name)
{
    char filetoparse[200];
    char *tmp = NULL;
    service_cache_h *got_serv = NULL;

    assert(process_name);
    /* example  INITNG_ROOT=/etc/initng process_name="deamon/sshd XML_EXT=".xml" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.xml" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.xml" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.xml" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.xml" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.xml */
    /* TRY NO 6, try load   "/etc/initng/sshd.xml" */
    /* TRY NO 7, try load   "/etc/initng/daemon.xml */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/default" XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 4, try "/etc/initng/daemon/default.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/default" XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 5, try "/etc/initng/daemon/daemon.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 6, try "/etc/initng/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 7, try "/etc/initng/daemon.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    D_("initng_ezxml_load_process(%s): got \"NULL\"\n", process_name);
    return (NULL);
}

/* parse a file for services */
static service_cache_h *initng_parse_ezxml_file(const char *filename,
                                                const char *watch_for)
{
    service_cache_h *service_to_return = NULL;
    service_cache_h *tmp = NULL;

    int file_fd;
    ezxml_t xml_file;

    assert(filename);
    assert(watch_for);
    D_("parse_xml_file(%s);\n", filename);


    if ((file_fd = open(filename, O_RDONLY, 0)) == -1)
    {
        D_("parse_ezxml_file(%s): Can't open xml file for parsing.\n",
           filename);
        return (NULL);
    }

    /* open xml file */
    if (!(xml_file = ezxml_parse_fd(file_fd)))
    {
        D_("parse_ezxml_file(%s): Can't parse file descriptor.\n", filename);
        close(file_fd);
        return (NULL);
    }

    /* big loop */
    while (xml_file)
    {
        stype_h *stype = NULL;

        D_("headnode: %s\n", xml_file->name);

        if ((stype = initng_service_types_get(xml_file->name)))
        {
            if ((tmp =
                 initng_parse_service_ezxml(xml_file, watch_for, NULL,
                                            stype)))
            {
                service_to_return = tmp;
            }

        }
        xml_file = xml_file->ordered;

    }


    /*exit_initng(1); */
    close(file_fd);
    if (service_to_return)
        return (service_to_return);
    return (NULL);
}


static service_cache_h *initng_parse_service_ezxml(ezxml_t node,
                                                   const char *watch_for,
                                                   service_cache_h * father,
                                                   stype_h * type)
{
    char *name = NULL;
    service_cache_h *copy = NULL;
    service_cache_h *new_service = NULL;
    service_cache_h *service_to_return = NULL;

    assert(watch_for);
    if (!node)
        return (NULL);

    D_("initng_parse_service_ezxml();\n");

    if (!node->attr[0] ||
        !strcasecmp(node->attr[0], "name") == 0 ||
        !(node->attr[1]) || !(name = i_strdup(node->attr[1])))
    {
        F_("initng_parse_service_ezxml(): service must have a name!");
        return (NULL);
    }

    D_("initng_parse_service_ezxml(): service name: \"%s\"\n", name);

    /* create a service entry */
    if (node->attr[2] &&
        strcmp(node->attr[2], "father") == 0 && node->attr[3])
    {
        if ((copy =
             initng_service_cache_find_by_name((const char *) node->attr[3])))
            new_service = initng_service_cache_copy(name, copy);
    }
    else
    {
        new_service = initng_service_cache_new(name, type);
    }

    if (!new_service)
    {
        F_("initng_parse_service_ezxml(): Unable to allocate service.\n");
        return (NULL);
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;

    node = node->child;

    while (node)
    {
        service_cache_h *tmp = NULL;
        stype_h *stype = NULL;

        /*
           if(strcmp((char *)node->name,"text")==0) {
           node=node->next;    
           continue;
           }
         */
        D_("node-name:%s\n", node->name);
        if ((stype = initng_service_types_get((const char *) node->name)))
        {
            if ((tmp =
                 initng_parse_service_ezxml(node->child, watch_for, NULL,
                                            stype)))
            {
                service_to_return = tmp;
            }
        }

        /* parse line - search for keywords in g.option_table */
        parse_xml_opt((char *) node->name, node->txt, stype, new_service);

        /* go to next node */
        node = node->ordered;
    }

    /* add to service db */
    if (!initng_service_cache_add(new_service))
    {
        F_("can't add service to service db!\n");
        initng_service_cache_free(new_service);
        return (NULL);
    }

    /* check if this is the service we ar looking for */
    if (service_match(watch_for, new_service->name) == TRUE)
        return (new_service);

    /* if we got a service from a under service */
    if (service_to_return)
        return (service_to_return);
    /* return null if not */
    return (NULL);

}

static int parse_xml_opt(const char *name, const char *value, stype_h * stype,
                         service_cache_h * srv)
{
    s_entry *current = NULL;

    while_service_data_types(current)
    {
        /* check that opt_name exist */
        if (!current->opt_name)
            continue;

        /* if it is bound to a special stype, check this */
        if (current->ot && current->ot != stype)
            continue;

        /* if option-name differs, continuue */
        if (strcmp(name, current->opt_name) != 0)
            continue;

        switch (current->opt_type)
        {
            case STRING:
                return (string_xml_parser(current, value, srv));
            case STRINGS:
                return (strings_xml_parser(current, value, srv));
            case SET:
                return (set_xml_parser(current, value, srv));
            case INT:
                return (int_xml_parser(current, value, srv));
            default:
                break;
        }
        break;
    }

    F_("parse_xml_opt(%s): error line: \"%s\"!\n", name, value);
    F_("sleeping 2 seconds......\n");
    sleep(2);
    return (FALSE);
}

static int set_xml_parser(s_entry * type, const char *value,
                          service_cache_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);

    D_("set_xml_parser(%i,%s,%s);\n", type, value, from_service->name);
    initng_service_cache_set(type, from_service);
    return (TRUE);
}


/* funtion copy full string into db */
static int string_xml_parser(s_entry * type, const char *value,
                             service_cache_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);
    D_("string_parser(%i,%s,%s);\n", type, value, from_service->name);
    initng_service_cache_set_string(type, from_service, i_strdup(value));
    return (FALSE);
}


static int int_xml_parser(s_entry * type, const char *value,
                          service_cache_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);
    D_("string_parser(%i,%s,%s);\n", type, value, from_service->name);
    initng_service_cache_set_int(type, from_service, atoi(value));
    return (FALSE);
}

/* funtion splits up all string to single entrys */
static int strings_xml_parser(s_entry * type, const char *value,
                              service_cache_h * from_service)
{
    char *to = NULL;
    char *tmp = (char *) value;

    assert(from_service);
    if (!value)
        return (FALSE);
    D_("strings_xml_parser(%i,%s,%s);\n", type, value, from_service->name);

    while ((to = st_dup_next_word(&tmp)))
    {
        inting_service_cache_set_another_string(type, from_service, to);
        JUMP_SPACES(tmp);
    }
    return (TRUE);
}






int module_init(void)
{
    D_("module_init();\n");
    initng_plugin_hook_add(&g.PARSERS, 10, &initng_ezxml_load_process);
    return (TRUE);
}

void module_unload(void)
{

    D_("module_unload();\n");
    initng_plugin_hook_del(&g.PARSERS, &initng_ezxml_load_process);

}
