/* Initng, a next generatind sysvinit replacement.
 * A plugin to launch services just before X starts.
 * The start of X is delayed until all jbX services are started.
 * Copyright (C) 2005 Thomas 'SaTaN0rX' Ilnseher <illth@gmx.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_static_states.h"

#include "initng_jbx.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <string.h>

typedef struct
{
    active_db_h *serv;
    const char *reqd_services;
    int req_wants_to_start;
} X_type;


static X_type X_services[100];
static int num_services = 0;

s_entry JBX = { "just_before", STRING, NULL,
    "The start of X is delayed until all jbX services are started."
};

/* s_entry PX = {  "provides_x", SET, 10, STD_OPT }; */

/*
 * as of now, this plugin only supports one X service (this isn't a limitation)
 * and ONE just_before_X service (this IS a limitation)
 */
static int check_X(active_db_h * service, a_state_h * status)
{
    int is_jbx;
    int is_X;

    int allow_to_start;

    int i;
    const char *X_serv_name;

    S_;

    /*    is_jbx = active_db_is(&JBX, service);
       is_X = active_db_is(&PX, service); */

    X_serv_name = initng_active_db_get_string(&JBX, service);

    is_jbx = (X_serv_name != NULL);

    if (is_jbx)
    {
        /* see if we already have an entry for this one */
        int add_to_list = TRUE;

        printf("service \"%s\" wants to start just before %s\n",
               service->name, X_serv_name);
        for (i = 0; i < num_services; i++)
        {
            if (X_services[i].serv == service)
            {
                /* found it! */
                add_to_list = FALSE;

                /* check if the associated req service wants to start ? */
                if (X_services[i].req_wants_to_start)
                {
                    /* the associated service wants to start => let this one start */
                    printf
                        ("allowing \"%s\" to start, as it's requirement wants to start\n",
                         service->name);
                    return (TRUE);
                }

                /* exit loop */
                i = num_services;
            }
        }
#ifdef DEBUG
        if (!add_to_list)
        {
            printf("it's requirement doesn't want to start\n");
        }
        else
        {
            printf("it's not in our list\n");
        }
#endif
        if (add_to_list && num_services < 100)
        {
            X_services[num_services].serv = service;
            X_services[num_services].reqd_services = X_serv_name;
            X_services[num_services].req_wants_to_start = FALSE;
            num_services++;
        }
    }

    allow_to_start = TRUE;
    is_X = FALSE;
    /* check if this is a requirement */
    for (i = 0; i < num_services; i++)
    {
        if (strstr(X_services[i].reqd_services, service->name))
        {
            printf("The current service \"%s\" is requrement to \"%s\"\n",
                   service->name, X_services[i].serv->name);
            is_X = TRUE;

            if (IS_UP(X_services[i].serv))
            {
                continue;
            }
            else
            {
                allow_to_start = FALSE;
                /* this one is a requirement to something,
                 * and something is NOT running */
            }

            /* check if X wants to start */
            if (status == &START_DEP_MET)
            {
                X_services[i].req_wants_to_start = TRUE;
                initng_global_set_sleep(1);
            }
        }
    }

    /* if the service neither has a just_before line, nor is a X service with unstart 
     * just_before deps */
    if (allow_to_start && !is_jbx)
        return (TRUE);

    /* only check START_DEP_MET */
    if (status != &START_DEP_MET)
        return (TRUE);

    /* 
     * here, the service:
     * a) wants to start
     * b) has a just_before= line
     * c) none of the mentioned just_before services does want to start
     */

    return (FALSE);
}


int module_init(const char *version)
{
    S_;
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&JBX);
    initng_active_state_add_test(&START_DEP_MET, 80, &check_X);
    return (TRUE);
}

void module_unload(void)
{
    S_;
    initng_service_data_types_del(&JBX);
    initng_active_state_del_test(&START_DEP_MET, &check_X);
}
