/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_env_variable.h"

#include "initng_logfile.h"

s_entry LOGFILE =
    { "logfile", STRING, NULL, "An extra output of service output." };


static int program_output(active_db_h * service, process_h * x)
{
    const char *filename = NULL;
    char *filename_fixed = NULL;
    int len = 0;
    int fd = -1;

    assert(service);
    assert(service->name);
    assert(x);

    D_("%s process fd: # %i, %i, service %s, have something to say\n",
       x->pt->name, x->out_pipe[0], x->out_pipe[1], service->name);

    /* get the filename */
    filename = initng_active_db_get_string(&LOGFILE, service);
    if (!filename)
    {
        D_("Logfile not set\n");
        return (FALSE);
    }

    /* Fix $variables in filename string */
    filename_fixed = fix_variables(filename, service);

    /* open the file */
    fd = open(filename, O_WRONLY | O_CREAT | O_APPEND);
    if (fd < 1)
    {
        F_("Error opening %s, err : %s\n", filename, strerror(errno));
        return (FALSE);
    }

    /* Write data to logfile */
    D_("Writing data...\n");
    len = strlen(&x->buffer[x->buffer_pos]);

    if (write(fd, &x->buffer[x->buffer_pos], len) != len)
        F_("Error writing to %s's log, err : %s\n", service->name,
           strerror(errno));

    free(filename_fixed);
    close(fd);

    return (TRUE);
}

int module_init(const char *version)
{

    S_;

    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&LOGFILE);

    initng_plugin_hook_add(&g.PIPEWATCHERS, 30, &program_output);
    return (TRUE);
}

void module_unload(void)
{
    S_;

    D_("module_unload();\n");

    initng_service_data_types_del(&LOGFILE);

    initng_plugin_hook_del(&g.PIPEWATCHERS, &program_output);
}
