
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include <errno.h>

/* #include "ngc.h" */
#include "../../src/initng.h"
#include "../../src/initng_global.h"
#include "../../src/initng_control_command.h"
#include "../../src/initng_active_state.h"
#include "../../src/initng_toolbox.h"
#include "../../initng-paths.h"

#include "initng_ngc2.h"

/* maximum size of string rescived, warning ngc -p can contain a lot */
#define MAX_RESCIVE_SIZE 100000

#define TRUE 1
#define FALSE 0
int debug = FALSE;

int header_printed = FALSE;
int try_count = 0;
int quiet = FALSE;

#define print_out(...); { if(quiet==FALSE) printf(__VA_ARGS__ ); }

#define TMP_QUIET if(quiet==FALSE) quiet=3;
#define TMP_UNQUIET if(quiet==3) quiet=FALSE;

/* Global comunication socket */
static int sock = -1;

/* functions */
void print_usage(void);
int open_socket(void);
int close_socket(void);
static result_desc *send_command(const char c, const char *l, const char *o);
void command_fetch_string(void);
int command_fetch_int(void);
int main(int argc, char *argv[]);
int ngc2_recvall(int sock, void *buf, int len);

int ngc_hlp(void);
int ngc_active_db(void);
int ngc_option_db(void);

void bad_syntax(void);
char *Argv;


#define FILL_SPACES(string, no) { \
    int t=0; \
    while(t<no) \
    { \
	if((string)[t]=='\0') \
	{ \
	    (string)[t]=' '; \
	    (string)[t+1]='\0'; \
	} \
	t++; \
    } \
}

/* print usage on screen */
void print_usage(void)
{
    send_command('h', NULL, NULL);
}

/* open socket to communicate with */
int open_socket()
{
    int len;
    struct sockaddr_un sockname;

    /* Create the socket. */
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0)
    {
        print_out("Failed to init socket\n");
        return (FALSE);
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;
    if (debug)
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_TEST);
    else
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_REAL);

    len += sizeof(sockname.sun_family);

    if (connect(sock, (struct sockaddr *) &sockname, len) < 0)
    {
        close(sock);
        print_out("Error connecting to initng socket\n");
        print_out("This probably means that you're not running initng.\n");
        exit(1);
        return (FALSE);
    }
    return (TRUE);
}


/* close pipes */
int close_socket(void)
{
    if (sock != -1)
    {
        close(sock);
        sock = -1;
    }
    return (sock);
}

static int send_and_handle(const char c, const char *l, const char *o);


/* send a command */
static result_desc *send_command(const char c, const char *l, const char *o)
{

    read_header header;
    result_desc *result = NULL;

    /* clear structure just in case */
    memset(&header, 0, sizeof(header));
    header.p_ver = PROTOCOL_VERSION;

    /* fill the header with data */
    header.c = c;
    if (l)
        strncpy(header.l, l, 100);
    else
        header.l[0] = '\0';
    if (o)
        strncpy(header.o, o, 200);
    else
        header.o[0] = '\0';


    /*print_out("Sending: %c, %s, %s\n", c, l ,o); */
    if ((sock < 0) && (open_socket() == FALSE))
    {
        print_out("Socket is not open, or failed to open!\n");
        exit(1);
    }

    if (write(sock, &header, sizeof(read_header)) <
        (signed) sizeof(read_header))
    {
        print_out("Could not send command!\n");
        exit(1);
    }

    /* no idea to stress initng */
    usleep(50000);

    /* if we are reloading initng, we wont have an answ from it, so don't fetch any */
    if (header.c == 'c')
    {
        print_out
            ("Can't give you feedback if reloading of initng succeeds or not.\n");
        exit(0);
    }

    /* FETCH THE RESULT */

    /* calloc the result space */
    result = calloc(1, sizeof(result_desc));

    /* read header data */
    if (TEMP_FAILURE_RETRY(read(sock, result, sizeof(result_desc))) !=
        sizeof(*result))
    {
        print_out("failed to fetch the result\n");
        exit(1);
    }

    /* check that protocol matches */
    if (result->p_ver != PROTOCOL_VERSION)
    {
        print_out("PROTOCOL_VERSION missmatch :\n");
        print_out
            (" initng_protocol_version: %i, client_protocol_version: %i !\n",
             result->p_ver, PROTOCOL_VERSION);
        print_out(" initng_version: %s, ngc version: %s\n", result->version,
                  VERSION);
        exit(1);
    }

    if (header_printed == FALSE)
    {
        /*print_out("Got an response from initng, version: %s\n", result->version); */

        if (strcmp(result->version, VERSION) != 0)
        {
            print_out("Version of initng and version of ngc differs!\n");
            print_out
                ("this might be a bad thing, that can in worst case cause problems!\n");
        }

        /*print_out("got an read: command %c, status %i, type %i\n", result->c,
           result->s, result->t); */
        header_printed = TRUE;
    }

    switch (result->s)
    {
        case S_FALSE:
            print_out("\nThe result of -%c is false = %i!\n\n", result->c,
                      result->s);
            free(result);
            return (NULL);

        case S_REQUIRES_OPT:
            print_out("\nThe command -%c requires an option.\n\n", result->c);
            free(result);
            return (NULL);

        case S_NOT_REQUIRES_OPT:
            print_out("\nThe command -%c must have no option.\n\n",
                      result->c);
            free(result);
            return (NULL);

        case S_INVALID_TYPE:
            print_out
                ("\nThe data returning of this command ( -%c ) is an unknown type.\n\n",
                 result->c);
            free(result);
            return (NULL);

        case S_COMMAND_NOT_FOUND:
            {
                print_out("\nCommand not found.\n\n");
                free(result);
                print_usage();
                return (NULL);
            }
        case S_TRUE:
            return (result);
    }
    return (NULL);
}

/* return the fetched string */
void command_fetch_string(void)
{
    char tmp_read[MAX_RESCIVE_SIZE + 1];
    ssize_t len;

    while ((len =
            TEMP_FAILURE_RETRY(read(sock, &tmp_read, MAX_RESCIVE_SIZE))) > 0)
    {
        tmp_read[len] = '\0';
        print_out("%s\n", tmp_read);
    }

    /* TODO: error handling??? */

    close_socket();
}

/* return the fetched return number */
int command_fetch_int(void)
{
    char got_from_read[20];

    got_from_read[read(sock, &got_from_read, 19)] = '\0';
    close_socket();
    return (atoi(got_from_read));
}

static int ngc_start_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;

    print_out("Starting service: \n");

    while (counter < 30)
    {
        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            print_out("\tFailed to fetch result\n");
            close_socket();
            return FALSE;
        }

        close_socket();
        /*print_out("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */
        if (row.i == IS_UP || row.i == IS_WAITING || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*print_out("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);

        /* make sure we got one */
        if (!result || result->c != 's')
        {
            print_out("\tInvalid result!\n");
            close_socket();
            return (FALSE);
        }

        counter++;
    }
    close_socket();

    if (row.i == IS_UP)
    {
        if (strcmp(row.s, "ALREDY_RUNNING") == 0)
        {
            print_out("\tService %s alredy running!\n", row.n);
        }
        else
        {
            print_out("\tService %s started sucessfully! (%s)\n", row.n,
                      row.s);
        }
        return (TRUE);
    }

    if (row.i == IS_WAITING)
    {
        print_out("\tService %s is starting! (%s)\n", row.n, row.s);
        return (TRUE);
    }

    if (counter == 10)
    {
        print_out
            ("Timeout on start service, might be started soon anyway.\n");
        return (FALSE);
    }

    if (row.i == IS_FAILED && try_count <= 2)
    {
        try_count++;
        print_out("\n\n");
        print_out("Warning, service %s is in status: \"%s\"\n\n", row.n,
                  row.s);

        if (strcmp(row.s, "NOT_FOUND") == 0)
        {
            print_out("This meens that the service cound not start,\n");
            print_out("Please check initng console, syslogs or ngc -L\n");
            print_out("For more info.\n\n");

            return (FALSE);
        }

        /* If found but failed */
        print_out("Service %s was failed, but found.\n", row.n);
        print_out
            ("Will not make an attempt to reset service state and try again\n");

        TMP_QUIET;                          /* dont print zap result */
        /* zap the service */
        if (!send_and_handle('z', NULL, row.n))
            return (FALSE);

        TMP_UNQUIET;


        sleep(1);
        print_out("\n\nTrying to start %s again ...\n\n", row.n);

        /* try start again */
        if (send_and_handle('u', NULL, row.n))
            return (TRUE);
    }

    print_out("\tFailed to start service, %s (%s)\n", row.n, row.s);
    return (FALSE);
}


static int ngc_stop_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;
    int once_found = 0;

    print_out("\nStopping service: \n");

    if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
        sizeof(row))
    {
        print_out("\tfailed to read response\n");
        close_socket();
        return FALSE;
    }

    close_socket();


    while (counter < 30)
    {
        /*print_out("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */

        if (row.i == IS_DOWN || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*print_out("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);


        /* make sure we got one */
        if (!result || result->c != 's')
        {
            print_out("\tInvalid result!\n");
            close_socket();
            return (FALSE);
        }

        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            print_out("\n\tfailed to read response\n");
            close_socket();
            return FALSE;
        }

        /* not found can be goood */
        if (once_found == 1 && strcmp(row.s, "NOT_FOUND") == 0)
            break;

        /* If its NOT_FOUND in future, then its a goot thing */
        once_found = 1;

        /* break if failed */
        if (row.i == IS_FAILED)
            break;

        close_socket();

        counter++;
    }
    close_socket();

    /* not found can be goood */
    if (once_found == 1 && strcmp(row.s, "NOT_FOUND") == 0)
    {
        print_out("\tService %s stopped sucessfully!\n", row.n);
        return (TRUE);
    }

    /* or bad */
    if (strcmp(row.s, "NOT_FOUND") == 0)
    {
        print_out("\n\n\tService %s was never started.\n", row.n);
        return (TRUE);
    }


    /* Down is really good to */
    if (row.i == IS_DOWN)
    {
        print_out("\tService %s is down!\n", row.n);
        return (TRUE);
    }


    print_out("Service %s has status failed to stop.\n", row.n);
    print_out("Instead you shud kill this service manually,\n");
    print_out("And after that do ngc -z %s, to release\n", row.s);
    print_out("its status from initng.\n\n");
    return (FALSE);
}


static int handle_result(result_desc * result)
{
    /* special handleing commands */
    switch (result->c)
    {
        case 'h':
            return (ngc_hlp());
        case 's':
            return (ngc_active_db());
        case 'O':
            return (ngc_option_db());
        case 'l':
            return (ngc_active_db());
        case 'u':
            return (ngc_start_result(result));
        case 'd':
            return (ngc_stop_result(result));
    }

    /* handle by type */
    switch (result->t)
    {
        case DATA_COMMAND:
            print_out("Uncatched data command!\n");
            bad_syntax();
            exit(1);

        case TRUE_OR_FALSE_COMMAND:
            {
                int r = command_fetch_int();

                if (r == TRUE)
                {
                    print_out(" command -%c returned SUCCESS! \n", result->c);
                }
                else
                {
                    print_out(" command -%c returned FAILURE! \n", result->c);
                }

                return (r);
            }
        case INT_COMMAND:
            print_out("INT_COMMAND_RESULT: %i\n", command_fetch_int());
            return (TRUE);
        case PRINT_COMMAND:
        case STRING_COMMAND:
            command_fetch_string();
            return (TRUE);
        case VOID_COMMAND:
            print_out("VOID_COMMAND\n");
            return (TRUE);
        case COMMAND_FAIL:
        default:
            close_socket();
            print_out("unkown data!, no haldler!\n");
            bad_syntax();
            exit(1);

    }
    print_out("Error, should not got here!\n report to bugzilla.\n");
    exit(22);
}

static int send_and_handle(const char c, const char *l, const char *o)
{
    result_desc *result = NULL;

    if (!(result = send_command(c, l, o)))
    {
        /*print_out("Command -%c --%s, %s - Unable to send command!\n", c, l, o); */
        return (FALSE);
    }
    if (!handle_result(result))
    {
        /*print_out("Command -%c --%s, %s - Bad result!\n", c, l, o); */
        return (FALSE);
    }

    if (result)
        free(result);
    /* if it is an -o option */
    return (TRUE);
}

/* THIS IS MAIN */
int main(int argc, char *argv[])
{
    int cc = 1;

    assert(argv[0]);

    header_printed = FALSE;
    Argv = argv[0];

    if (argv[1] && strcmp(argv[1], "--quiet") == 0)
        quiet = TRUE;

    /* print banner */
    print_out("Next Generation init Control. version: %s\n", VERSION);
    print_out("written by Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");



    /* check if cmd-line contains "ngdc", if set debug */
    if (strstr(argv[0], "ngdc"))
    {
        print_out("Warning. This is ngdc!\n");
        debug = TRUE;
    }


    if (debug == FALSE && getuid() != 0)
    {
        print_out("You need root access to communicate with initng.\n");
        exit(2);
    }

    /* make sure there are any arguments at all */
    if (argc < 2)
        print_usage();


    while (argv[cc])
    {
        char *opt = NULL;

        /* every fresh start needs a '-' char */
        if (argv[cc][0] != '-')
        {
            bad_syntax();
            exit(1);
        }

        /* check that there is a char after the '-' */
        if (!argv[cc][1])
        {
            bad_syntax();
            exit(1);
        }

        /* if next option contains data, but not starting with an '-', its considered an option */
        if (argv[cc + 1] && argv[cc + 1][0] != '-')
            opt = argv[cc + 1];


        /* if it is an --option */
        if (argv[cc][1] == '-')
        {
            if (!send_and_handle('\0', &argv[cc][2], opt))
                exit(1);
        }
        else
        {
            if (!send_and_handle(argv[cc][1], NULL, opt))
                exit(1);
        }

        cc++;
        if (opt)
            cc++;


    }
    printf("\n\n");
    exit(0);
}

int ngc2_recvall(int sock, void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = recv(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 0;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 1;
}


/* get help */
int ngc_hlp(void)
{
    /* data got */
    help_row row;

    /* print head */
    print_out(" ngc understand this commands:\n");
    print_out(" option                          : description\n");
    print_out
        (" ----------------------------------------------------------\n");

    while (ngc2_recvall(sock, &row, sizeof(help_row)))

    {
        int i;
        char lname[32];

        /* copy name to the new static array */
        strncpy(lname, row.l, 25);
        lname[25] = '\0';

        switch (row.o)
        {
            case USES_OPT:
                strcat(lname, " <opt>");
                break;
            case REQUIRES_OPT:
                strcat(lname, " opt");
                break;
            default:
                break;
        }

        /* Fill out 40 blanks for nice output */
        for (i = strlen(lname); i <= 31; i++)
            lname[i] = ' ';

        lname[31] = '\0';
        if (row.c != '\0')
        {
            print_out(" [-%c] --%s: %s\n", row.c, lname, row.d);
        }
        else
        {
            print_out("       --%s: %s\n", lname, row.d);
        }
    }
    return (TRUE);
}


int ngc_active_db(void)
{
    active_row row;
    struct timeval now;

    gettimeofday(&now, NULL);

    /* print head */
    print_out
        ("hh:mm:ss service                                :status                   since\n");
    print_out
        ("--------------------------------------------------------------------------------\n");

    while (ngc2_recvall(sock, &row, sizeof(active_row)))
    {
        struct tm *ts = localtime(&row.t.tv_sec);
        int tbuf = MS_DIFF(now, row.t) / 1000;

        print_out("%.2i:%.2i:%.2i %-38s :%-22s %7i%c\n",
                  ts->tm_hour, ts->tm_min, ts->tm_sec, row.n, row.s,
                  tbuf > 60 ? tbuf / 60 : tbuf, tbuf > 60 ? 'm' : 's');

    }
    close_socket();
    return (TRUE);
}


int ngc_option_db(void)
{
    option_row row;
    char ct[20];

    /* print head */
    print_out(" %-10s %-24s = %-8s %s\n", "WhereToUSE", "Name", "ValueToUSE",
              "Description");
    print_out
        (" ----------------------------------------------------------------\n");

    while (ngc2_recvall(sock, &row, sizeof(option_row)))
    {
        switch (row.t)
        {
            case STRING:
                strcpy(ct, "STRING");
                break;
            case STRINGS:
                strcpy(ct, "STRINGS");
                break;
            case SET:
                strcpy(ct, "SET");
                break;
            case INT:
                strcpy(ct, "INT");
                break;
            case ALIAS:
                strcpy(ct, "ALIAS");
                break;
            default:
                strcpy(ct, "U_D_T");
                break;
        }

        print_out(" %-10s %-24s = %-8s %s\n", row.o, row.n, ct, row.d);

    }
    close_socket();
    return (TRUE);
}


/* Avoid using goto */
void bad_syntax(void)
{
    print_out("Bad syntax!\n\n");
    print_usage();
}
