
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdio.h>
#include <string.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_static_process_types.h"
#include <assert.h>

#include "initng_pause.h"
s_entry S_DELAY = { "exec_delay", VARIABLE_INT, NULL,
    "Pause this much seconds before service is launched."
};

s_entry MS_DELAY = { "exec_m_delay", VARIABLE_INT, NULL,
    "Pause this much microseconds before service is launched."
};

int do_pause(active_db_h * s, process_h * p)
{
    int s_delay = 0;
    int ms_delay = 0;

    assert(s);
    assert(p);
    assert(p->pt);

    D_(" %s\n", s->name);


    s_delay = initng_active_db_get_int_var(&S_DELAY, p->pt->name, s);
    ms_delay = initng_active_db_get_int_var(&MS_DELAY, p->pt->name, s);


    if (s_delay)
    {
        D_("Sleeping for %i seconds.\n", s_delay);
        sleep(s_delay);
    }

    if (ms_delay)
    {
        D_("Sleeping for %i miliseconds.\n", ms_delay);
        usleep(ms_delay);
    }

    return (TRUE);
}

int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }
    initng_service_data_types_add(&S_DELAY);
    initng_service_data_types_add(&MS_DELAY);
    return (initng_plugin_hook_add(&g.A_FORK, 90, &do_pause));
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_service_data_types_del(&S_DELAY);
    initng_service_data_types_del(&MS_DELAY);
    initng_plugin_hook_del(&g.A_FORK, &do_pause);
}
