/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_service_cache.h"

#include "initng_provide.h"

static int virtual_add(char *name, a_status_h * status, char *father);
static int virtual_del(char *name);
static int virtual_addual_up(char *arg);


s_entry PROVIDE = { "provide", STRINGS, STD_OPT,
    "A name of a virtual service that is created when this service is up."
};
s_command SET_VIRTUAL_UP = { 'V', "set_virtual_up", TRUE_OR_FALSE_COMMAND,
    REQUIRES_OPT,
    {(void *) &virtual_addual_up}, "Set a virtual"
};

/* returns TRUE if all use deps are started */
int check_provide(active_db_h * service)
{
    char *tmp = NULL;

    assert(service);
    assert(service->name);

    switch (service->a_status)
    {
        case LOADING:
            while ((tmp =
                    initng_active_db_get_next_string(&PROVIDE, service, tmp)))
            {
                virtual_add(tmp, LOADING, service->name);
            }
        case RUNNING:
        case DONE:
            while ((tmp =
                    initng_active_db_get_next_string(&PROVIDE, service, tmp)))
            {
                virtual_add(tmp, DONE, service->name);
            }
            break;
        case STOPPING:
        case STOPPED:
            while ((tmp =
                    initng_active_db_get_next_string(&PROVIDE, service, tmp)))
            {
                virtual_del(tmp);
            }
            break;
        default:
            break;
    }
    return (TRUE);
}

static int virtual_add(char *name, a_status_h * status, char *father)
{
    active_db_h *act = NULL;

    if (!name)
    {
        F_("Where is the name!\n");
        return (FALSE);
    }

    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        printf("System is in wrong system_state for this!\n");
        return (FALSE);
    }

    act = initng_active_db_find_by_exact_name(name);
    if (!act)
    {
        if (!(act = initng_active_db_new(i_strdup(name))))
        {
            F_("Failed to create virtual: %s\n", name);
            return (FALSE);
        }
        if (initng_active_db_add(act) != TRUE)
        {
            F_("Failed to add %s\n", act->name);
            initng_active_db_free(act);
            return (FALSE);
        }
        act->type = TYPE_VIRTUAL;
    }
    initng_common_mark_service(act, status);
    if (father)
        initng_active_db_set_another_string(&NEED, act, i_strdup(father));
    return (TRUE);
}

static int virtual_del(char *name)
{
    active_db_h *act = NULL;

    assert(name);

    /* check if an virtual exits - then set its status */
    if ((act = initng_active_db_find_in_name(name))
        && (act->type == TYPE_VIRTUAL))
    {
        initng_common_mark_service(act, STOPPED);
        initng_active_db_del(act);
        initng_active_db_free(act);
        act = NULL;
        return (TRUE);
    }
    return (FALSE);
}
static int virtual_addual_up(char *arg)
{
    return (virtual_add(arg, DONE, NULL));
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&PROVIDE);
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 80, &check_provide);
    initng_command_add(&SET_VIRTUAL_UP);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_service_data_types_del(&PROVIDE);
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &check_provide);
    initng_command_del(&SET_VIRTUAL_UP);
}
