
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>

#include "../../src/initng_active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_execute.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_common.h"
#include "../../src/initng_signal.h"
#include "../../src/initng_fork.h"
#include "../../src/initng_env_variable.h"

#include "initng_simple_launcher.h"

/*#undef D_
   #define D_ W_ */

s_entry EXEC =
    { "exec", VARIABLE_STRING, NULL,
"Contains the path and argumetns to a file to exec." };

s_entry EXECS = { "exec_path", VARIABLE_STRINGS, NULL,
    "The path for one ore more executables."
};
s_entry EXEC_ARGS = { "exec_args", VARIABLE_STRING, NULL,
    "The arguments for the executable."
};

/*#ifdef DEBUG
   static void D_argv(const char *o, char **argv)
   {
   int i;

   if (!argv)
   return;

   for (i = 0; argv[i]; i++)
   D_("%s[%-2i]: %s\n", o, i, argv[i]);
   }
   #endif */

/*
 * Searches for exec in PATH.
 * 
 * Completely rewritten function.
 * Contains old Code from SaTaN0rX and DEac-.
 * @author TheLich
 * @param exec filename, which searched.
 * @return executeable file with absolute path.\nIf exec was absolute, this will be a pointer to exec.
 * On failure, it will return NULL.
 */

static char *expand_exec(char *exec)
{
    char *filename = NULL;
    size_t exec_len = 0;
    size_t path_c, i, len = 0;
    char *PATH = NULL;
    char **path_argv = NULL;
    struct stat test;


    if (!exec)
        return NULL;

    /* 11/20/2005 SaTaN0rX: preliminary support for searching the PATH
     * only search the PATH if this is NOT already an absolute path
     */

    /* if path provided, youst return */
    if (exec[0] == '/')
        return (NULL);

    /* fet exec string lenght to later use */
    exec_len = strlen(exec);

    D_("initng_s_launch: %s is not an absolute path, searching $PATH\n",
       exec);

    /* get the env-path variable */
    PATH = getenv("PATH");

    /* Make sure we got a path */
    if (!PATH)
    {
        D_("No $PATH found, using default path\n");
        PATH =
            i_strdup
            ("/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin");
    }
    else
    {
        /* PATH will later be changed, so we have to use a duplicate */
        PATH = i_strdup(PATH);
    }

    D_("PATH determined to be %s\n", PATH);

    /* split path by ':' char */
    path_argv = split_delim(PATH, ":", &path_c, 0);

    /* walk the list of entrys */
    for (i = 0; path_argv[i]; i++)
    {
        /* get entry lenght */
        len = strlen(path_argv[i]);

        /* what does this do? */
        filename = (char *) i_calloc(exec_len + len + 2, sizeof(char));
        strcpy(filename, path_argv[i]);
        if (filename[len - 1] != '/')
            strcat(filename, "/");
        strcat(filename, exec);

        /* check so that file exits, if exists leave this loop */
        if ((stat(filename, &test) != -1)
            && (test.
                st_mode & (S_IXOTH | S_IXGRP | S_IXUSR | S_ISVTX | S_ISGID |
                           S_ISUID | S_IFREG)))
            break;

        /* cleanup */
        free(filename);
        filename = NULL;
    }

    /* free */
    free(PATH);
    free(path_argv);
    PATH = NULL;
    path_argv = NULL;

    /* make sure we got a filename to use */
    if (!filename)
        F_("No executable found\n");

    /* return the filename */
    return filename;
}

static int simple_exec_fork(process_h * process_to_exec, active_db_h * s,
                            size_t argc, char **argv)
{
    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */

    if ((pid_fork = initng_fork(s, process_to_exec)) == 0)
    {

#ifdef DEBUG
        D_("FROM_FORK simple_exec(%i,%s, ...);\n", argc, argv[0]);
        /*D_argv("simple_exec: ", argv); */
#endif

        /* FINALLY EXECUTE *//* execve replaces the running process */
        execve(argv[0], argv, new_environ(s));

        /* Will never get here if execve succeeded */
        F_("EEEEEEEEEEEEEEEEEEERRRRRRRRRRRRRRRRRRRRRRRRRRROOOOOOOOOOOOOOOOOOOOOOOOOORRRRRRRRRRRRRRR!!!!!!!!!\n");
        F_("Can't execute source %s!\n", argv[0]);

        _exit(1);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);

    if (pid_fork > 0)
    {
        process_to_exec->pid = pid_fork;
        return (TRUE);
    }

    process_to_exec->pid = 0;
    return FALSE;
    /* if to test want to lock this up util fork is done ...   waitpid(pid_fork,0,0); */
}                                           /* end fork_and_exec() */


static int simple_exec_try(char *exec, active_db_h * service,
                           process_h * process)
{
    const char *exec_args_unfixed = NULL;
    char **argv = NULL;
    size_t argc = 0;

    D_("exec: %s, service: %s, process: %s\n", exec, service->name,
       process->pt->name);

    /* exec_args should be parsed at the moment, too */
    exec_args_unfixed =
        initng_active_db_get_string_var(&EXEC_ARGS, process->pt->name,
                                        service);
    if (exec_args_unfixed)
    {
        char *exec_args = NULL;


        /* get some fixed variables, with ${VARIABLES} fixed, be aware that this is a new malloc and needs to be free() */
        exec_args = fix_variables(exec_args_unfixed, service);
        if (!exec_args)
        {
            F_("Failed to fix_variables: \"%s\"\n", exec_args);
            return (FALSE);
        }

        /* split the string, with entrys to an array of strings */
        argv = split_delim(exec_args, WHITESPACE, &argc, 1);

        /* make sure it succeded */
        if (!argv || !argv[0])
        {
            F_("split_delim exec_args returns NULL.\n");
            free(exec_args);
            exec_args = NULL;
            return (FALSE);
        }


    }
    else
    {
        /* we need a empty argv anyway */
        argv = (char **) i_calloc(2, sizeof(char *));
        argv[1] = NULL;
        argc = 0;
    }

    argv[0] = exec;

    return (simple_exec_fork(process, service, argc, argv));

}

static int simple_exec(active_db_h * service, process_h * process)
{
    const char *exec = NULL;
    struct stat stat_struct;

    D_("service: %s, process: %s\n", service->name, process->pt->name);

    while ((exec =
            initng_active_db_get_next_string_var(&EXECS, process->pt->name,
                                                 service, exec)))
    {
        int res = FALSE;
        char *exec_fixed = NULL;

        /* be aware that fix_variables() return is a malloc, and needs to be free */
        exec_fixed = fix_variables(exec, service);

        if (!exec_fixed)
        {
            F_("Unable to fix_variables!\n");
            return (FALSE);
        }

        /* check if the file exist */
        if (stat(exec_fixed, &stat_struct) != 0)
        {
            D_(" note, %s exec_fixed does not exist. \n", exec_fixed);
            continue;
        }

        /* Try to execute that one */
        res = simple_exec_try(exec_fixed, service, process);

        /* cleanup */
        free(exec_fixed);
        exec_fixed = NULL;

        /* Return true if sucessfull */
        if (res == TRUE)
            return (TRUE);

    }

    return (FALSE);
}

static int simple_run(active_db_h * service, process_h * process)
{
    const char *exec = NULL;
    char *exec_fixed = NULL;
    char **argv = NULL;
    size_t argc = 0;
    int result = FALSE;
    char *argv0 = NULL;

    D_("service: %s process: %s.\n", service->name, process->pt->name);

    exec = initng_active_db_get_string_var(&EXEC, process->pt->name, service);
    if (!exec)
        return (FALSE);


    /* be aware that fix_variables() return is a malloc, and needs to be free */
    exec_fixed = fix_variables(exec, service);
    if (!exec_fixed)
    {
        F_("Unable to fix_variables!\n");
        return (FALSE);
    }

    /* argv-entries are pointer to exec_t[x] */
    argv = split_delim(exec_fixed, WHITESPACE, &argc, 0);

    /* make sure we got something from the split */
    if (!argv || !argv[0])
    {
        D_("split_delim on exec returns NULL.\n");
        free(exec_fixed);
        exec_fixed = NULL;
        return (FALSE);
    }

    /* if it not contains a full path */
    if (argv[0][0] != '/')
    {
        argv0 = expand_exec(argv[0]);
        if (!argv0)
        {
            D_("%s was not found in search path.\n", argv[0]);
            free(argv);
            argv = NULL;
            free(exec_fixed);
            exec_fixed = NULL;
            return (FALSE);
        }
        argv[0] = argv0;
    }


    /* try to execute, remember the result */
    result = simple_exec_fork(process, service, argc, argv);

    /* clean up */
    free(argv);
    argv = NULL;
    free(exec_fixed);
    exec_fixed = NULL;
    if (argv0)
    {
        free(argv0);
        argv0 = NULL;
    }

    /* return result */
    if (result == FAIL)
        return (FALSE);
    return (result);
}

static int initng_s_launch(active_db_h * service, process_h * process)
{

    assert(service);
    assert(service->name);
    assert(process);

    D_("service: %s, process: %s\n", service->name, process->pt->name);

    if (initng_active_db_is_var(&EXECS, process->pt->name, service))
        return (simple_exec(service, process));

    if (initng_active_db_is_var(&EXEC, process->pt->name, service))
        return (simple_run(service, process));

    return (FALSE);
}



int module_init(const char *version)
{
    D_("initng_simple_plugin: module_init();\n");

    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_plugin_hook_add(&g.LAUNCH, 40, &initng_s_launch);
    initng_service_data_types_add(&EXEC);
    initng_service_data_types_add(&EXECS);
    initng_service_data_types_add(&EXEC_ARGS);
    return (TRUE);
}

void module_unload(void)
{
    D_("initng_simple_plugin: module_unload();\n");

    initng_service_data_types_del(&EXEC);
    initng_service_data_types_del(&EXECS);
    initng_service_data_types_del(&EXEC_ARGS);
    initng_plugin_hook_del(&g.LAUNCH, &initng_s_launch);

}
