
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "initng.h"
#include "initng_global.h"
#include "initng_active_state.h"
#include "initng_toolbox.h"


int initng_active_state_add(a_state_h * state)
{
    assert(state);

    /* look for duplets */
    if (initng_active_state_find(state->state_name) != NULL)
    {
        F_("There exists a state with this state_name already, please check this!\n");
        return (FALSE);
    }

    D_("adding %s.\n", state->state_name);
    /* add this state, to the big list of states */
    list_add(&(state->list), &(g.states.list));

    /* initziate the list of tests for this state */
    INIT_LIST_HEAD(&state->test.list);

    /* return happily */
    return (TRUE);
}

a_state_h *initng_active_state_find(const char *state_name)
{
    a_state_h *current = NULL;

    assert(state_name);

    /* walk the state db */
    while_active_states(current)
    {
        if (strcmp(state_name, current->state_name) == 0)
            return (current);
    }
    return (NULL);
}

void initng_active_state_add_hook(a_state_h * state, int order, void *test,
                                  void *run)
{
    a_state_hook *current = NULL;
    a_state_hook *new = i_calloc(1, sizeof(a_state_hook));

    if (!new)
        return;

    new->state_test = test;
    new->state_run = run;

    /* if list is empty, add it as first */
    if (list_empty(&state->test.list))
    {
        list_add(&new->list, &state->test.list);
        return;
    }

    /* other walk the list, skip the lower orders */
    while_active_state_hooks(current, state)
    {
        if (current->order < order)
            continue;
    }

    /* add after that */
    list_add(&new->list, &current->list);
}

void initng_active_state_del_hook(a_state_h * state, void *hook)
{
    a_state_hook *current, *safe = NULL;

    while_active_state_hooks_safe(current, safe, state)
    {
        if (current->state_test == hook || current->state_run == hook)
        {
            list_del(&current->list);
            free(current);
        }
    }
}
