
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef ACTIVE_STATE_H
#define ACTIVE_STATE_H

typedef struct a_state_t a_state_h;

#include "initng_list.h"
#include "initng_active_db.h"

typedef enum
{
    IS_UNKNOWN = 0,
    IS_UP = 1,
    IS_DOWN = 2,
    IS_FAILED = 3,
    IS_STARTING = 4,
    IS_STOPPING = 5,
    IS_WAITING = 6
} e_is;

/* This is a test struct that can be added */
typedef struct
{
    /*
     * Function that will be called, before a serice is set this state.
     * If return is FALSE, it wont allow this change.
     */
    int (*state_test) (active_db_h * service, a_state_h * state);

    /*
     * This function will be run, when a service is set this state.
     */
    void (*state_run) (active_db_h * service);

    /*
     * the order of tests and runs.
     */
    int order;

    /* the list of tests */
    struct list_head list;
} a_state_hook;

struct a_state_t
{
    /* the name of the state in a string, will be printed */
    const char *state_name;

    /* If this state is set for a service, is it roghly: */
    e_is is;

    /*
     * This function will be rune on service with this state set, if g.interrupt is set.
     */
    void (*state_interrupt) (active_db_h * service);


    /*
     * The tests that shud be run, before a service can get this state.
     */
    a_state_hook test;

    /* The list this struct is in */
    struct list_head list;
};

int initng_active_state_add(a_state_h * state);


#define initng_active_state_del(st) list_del(&(st)->list)

#define initng_active_state_del_test(st, ho) initng_active_state_del_hook(st, ho)
#define initng_active_state_del_run(st, ho) initng_active_state_del_hook(st, ho)

#define initng_active_state_add_test(state, order, test) initng_active_state_add_hook(state, order, test, NULL)
#define initng_active_state_add_run(state, order, run) initng_active_state_add_hook(state, order, NULL, run)
void initng_active_state_add_hook(a_state_h * state, int order, void *test,
                                  void *run);
void initng_active_state_del_hook(a_state_h * state, void *hook);

a_state_h *initng_active_state_find(const char *state_name);

#define while_active_states(current) list_for_each_entry_prev(current, &g.states.list, list)
#define while_active_states_safe(current) list_for_each_entry_prev_safe(current, safe, &g.states.list, list)

#define while_active_state_hooks(current, state) list_for_each_entry_prev(current, &state->test.list, list)
#define while_active_state_hooks_safe(current, safe, state) list_for_each_entry_prev_safe(current, safe, &state->test.list, list)


#define IS_MARK(serv, state) (serv && serv->current_state && serv->current_state == state)
#define IS_UP(serv) (serv && serv->current_state && serv->current_state->is == IS_UP)
#define IS_DOWN(serv) (!serv->current_state || (serv && serv->current_state && serv->current_state->is == IS_DOWN))
#define IS_FAILED(serv) (serv && serv->current_state && serv->current_state->is == IS_FAILED)
#define IS_STARTING(serv) (serv && serv->current_state && serv->current_state->is == IS_STARTING)
#define IS_STOPPING(serv) (serv && serv->current_state && serv->current_state->is == IS_STOPPING)
#define IS_WAITING(serv) (serv && serv->current_state && serv->current_state->is == IS_WAITING)

#endif
