/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng_plist.h"
#include <assert.h>

int plist_add(void *pointer, plist_h * plist)
{
    int i;
    void **new_plist = NULL;

    assert(plist);
    assert(plist->alloc_heap);

    /* check if there is empty slots */
    for (i = 0; i < plist->allocated; i++)
    {
        if (plist->list[i] == NULL)
        {
            /* set the pointer */
            plist->list[i] = pointer;

            /* increase the limiting variable if this is last */
            if (i == plist->next_to_set)
                (plist->next_to_set)++;

            return (TRUE);
        }
    }

    /* set i to last entry */
    i = plist->allocated;

    /* if no empty space was found, we have to bigger the array */
    plist->allocated += plist->alloc_heap;
    new_plist = realloc(plist->list, (sizeof(void *) * plist->allocated));

    /* make sure it succeded */
    if (!new_plist)
        return (FALSE);

    /* then set it */
    plist->list = new_plist;

    /* zero the list */
    while (i < plist->allocated)
    {
        plist->list[i] = NULL;
        i++;
    }

    /* set the pointer to the last set one */
    plist->list[plist->next_to_set++] = pointer;

    return (TRUE);
}

int plist_del(void *pointer, plist_h * plist)
{
    int i;

    assert(plist);
    assert(plist->alloc_heap);

    /* search the list for matching entry */
    for (i = 0; i < plist->allocated; i++)
    {
        /* if not what we are looking for, continue walking */
        if (plist->list[i] != pointer)
            continue;

        /* null that entry */
        plist->list[i] = NULL;

        /* START CLEANING */

        /* while this is the last one, increase the size. */
        while (i + 1 == plist->next_to_set && plist->list[i] == NULL
               && plist->next_to_set != 0)
        {
            (plist->next_to_set)--;
            if (i > 0)
                i--;
        }

        /* check if its empty */
        if (plist->next_to_set == 0)
        {
            free(plist->list);
            plist->list = NULL;
            plist->allocated = 0;
            return (TRUE);
        }

        /* realloc if to big diff */
        if ((plist->allocated - plist->next_to_set) > plist->alloc_heap)
        {
            plist->allocated -= plist->alloc_heap;
            plist->list = realloc(plist->list,
                                  (sizeof(void *) * plist->allocated));
        }

        return (TRUE);

    }

    return (FALSE);
}

void plist_clear(plist_h * plist)
{
    if (plist->list)
    {
        free(plist->list);
    }
    plist->list = NULL;
    plist->allocated = 0;
    plist->next_to_set = 0;
}
