
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <fnmatch.h>

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <stdlib.h>
#include "initng_string_tools.h"
#include "initng_toolbox.h"

/*
 * this splits an long string, with separators, like \n , ; ' '
 * into an char **
 * buf: the line to split.
 * len; An pointer sent back with lenght
 * plus: Allocate this mutch extra char ** size;
 * ofs: (offset) start a bit into the buf
 *
 * OBS, this functions only creats a pointer list, that links
 * to buf, if buf is freed, the list dies!
 */
char **split_n(char *buf, int *len, int plus, int ofs)
{
    int n = 1;
    char **v = 0;
    char **w;

    /* step 1: count tokens */
    char *s;

    assert(buf);

    for (s = buf; *s; s++)
        if (*s == '\n' || *s == ',' || *s == ';' || *s == ' ')
            n++;
    /* step 2: allocate space for pointers */
    v = (char **) i_calloc((n + plus), sizeof(char *));
    if (!v)
        return 0;
    w = v + ofs;
    *w++ = buf;
    for (s = buf;; s++)
    {
        while (*s && *s != '\n' && *s != ',' && *s != ';' && *s != ' ')
            s++;
        if (*s == 0)
            break;
        if (*s == '\n' || *s == ',' || *s == ';' || *s == ' ')
        {
            *s = 0;
            *w++ = s + 1;
        }
    }
    *len = w - v;
    return v;
}


/** Relinks a string to a argv-like string-array.
 * Example: \t  /bin/executeable --ham -flt   --moohoho  lalala
 *       => \t  /bin/executeable\0--ham\0-flt\0  --moo\0hoho\0 lalala
 *    argv:     ^[0]              ^[1]   ^[2]    ^[3]   ^[4]   ^[5]
 *    argc: 6
 *
 * string: String to split.
 * delim : Char that splits the strings.
 * argc  : Pointer, sets the number of splits done.
 * ofs   : Offset, start from this arg to fill.
 * @idea DEac-
 * @author TheLich
 * @got_it_working powerj
 */
char **split_delim(char *string, const char *delim, size_t * argc, int ofs)
{
    int len;
    char **array = (char **) i_calloc(1, sizeof(char *));
    size_t i = 0;

    if (!string)
        return (NULL);

    while (string[ofs] != '\0')
    {
        len = strcspn(string + ofs, delim);
        if (len != 0)
        {
            i++;
            array = (char **) i_realloc(array, sizeof(char *) * (i + 1));
            array[i - 1] = string + ofs;
            if (string[len + ofs] != '\0')
            {
                string[len + ofs] = '\0';
                len++;
            }
        }
        else
            len = 1;

        ofs += len;
    }

    array[i] = NULL;
    *argc = i;
    return array;
}


/*
 * a simple strncasecmp, thats only returns TRUE or FALSE
 * it will also add the pointer of string, the value, to forward the
 * pinter to next word.
 */
int st_cmp(char **string, const char *to_cmp)
{
    int chars = 0;

    assert(to_cmp);
    assert(string);

    chars = strlen(to_cmp);

    /* skip beginning first spaces */
    JUMP_SPACES(*string)
                                                                                                        /* this might be an "comp pare" */
                                                                                                        if
                                                                                                        ((*string)[0] == '"' && to_cmp[0] != '"')
        (*string)++;

    /* ok, strcasecmp this */
    if (strncasecmp((*string), to_cmp, chars) == 0)
    {
        /* increase the string pointer */
        (*string) += chars;
        return (TRUE);
    }
    return (FALSE);
}

/*
 * Senses what in the string, that is a word, (can be a
 * line with "this is one word", it will i_strdup it, and
 * add word lenght on string pointer
 */
char *st_dup_next_word(char **string)
{
    char *td = NULL;
    int i = 0;

    assert(string);

    /* skip beginning first spaces */
    JUMP_SPACES(*string);
    if (!(*string)[0] || (*string)[0] == '\n')
        return (NULL);

    /* handle '"' chars */
    if ((*string)[0] == '"')
    {
        (*string)++;
        i = strcspn(*string, "\"");
        if (i < 1)
            return (NULL);
        td = i_strndup(*string, i);
        (*string) += i;
        if (*string[0] == '"')
            (*string)++;
        return (td);
    }

    /* handle '{' '}' chars */
    if ((*string)[0] == '{')
    {
        (*string)++;
        i = strcspn(*string, "}");
        if (i < 1)
            return (NULL);
        td = i_strndup(*string, i);
        (*string) += i;
        if (*string[0] == '}')
            (*string)++;
        return (td);
    }

    /* or copy until space tab newline, ; or , */
    i = strcspn(*string, " \t\n;,");
    if (i < 1)
        return (NULL);
    /* copy string */
    td = i_strndup(*string, i);
    (*string) += i;
    return (td);
}

/*
 * This will dup, in the string, what is resolved as the line,
 * (stops with '\n', ';', '\0'), copy it and increment *string with len
 */
char *st_dup_line(char **string)
{
    char *td = NULL;
    int i = 0;

    assert(string);

    /* skip beginning first spaces */
    JUMP_SPACES(*string);
    if (!(*string)[0] || (*string)[0] == '\n')
        return (NULL);

    /* handle '"' chars */
    if ((*string)[0] == '"')
    {
        (*string)++;
        i = strcspn(*string, "\"");
        if (i < 1)
            return (NULL);
        td = i_strndup(*string, i);
        (*string) += i;
        if (*string[0] == '"')
            (*string)++;
        return (td);
    }

    /* handle '{' '}' chars */
    if ((*string)[0] == '{')
    {
        (*string)++;
        i = strcspn(*string, "}");
        if (i < 1)
            return (NULL);
        td = i_strndup(*string, i);
        (*string) += i;
        if (*string[0] == '}')
            (*string)++;
        return (td);
    }

    /* or copy until space tab newline, ; or , */
    i = strcspn(*string, "\n;");
    if (i < 1)
        return (NULL);
    /* copy string */
    td = i_strndup(*string, i);
    (*string) += i;
    return (td);
}


/* function strips "test/go/th" to "th" */
const char *st_strip_path(const char *string)
{
    int i = 0;

    assert(string);

    /* go to end */
    while (string[i] != '\0')
        i++;

    /* go back to the '/' char */
    while (i && string[i] != '/')
        i--;

    if (string[i] == '/')
        return (&string[i + 1]);
    return (string);
}

/* function strips "test/go/th" to "test/go" */
char *st_get_path(const char *string)
{
    int i = 0;

    assert(string);

    /* go to end */
    while (string[i] != '\0')
        i++;
    /* step back */
    while (i && string[i] != '/')
        i--;

    if (i)
        return (i_strndup(string, i));

    return (i_strdup(string));
}

/* function strips "test/go/th" to "test/go" and then to "test" */
int st_strip_end(char **string)
{
    int i = 0;

    assert(string);

    /* stand on last char */
    while ((*string)[i] != '\0')
        i++;

    /* rewind util standing on a '/' */
    while (i && (*string)[i] != '/')
        i--;

    /* if we really are standing on a '/' */
    if ((*string)[i] == '/')
    {
        (*string)[i] = '\0';                /* truncate */
        return (TRUE);
    }
    return (FALSE);
}

/*
 * This search for a pattern.
 * If string is "net/eth0" then pattern "net/et*" and pattern "net/eth?"
 * shoud return true.
 */
int service_match(const char *string, const char *pattern)
{
    char *service_mask = NULL;

    assert(string);
    assert(pattern);
    D_("matching string: \"%s\", to pattern: \"%s\"\n", string, pattern);

    /* do pattern matching only if service name does not contain wildcards */
    if (strchr(string, '*') || strchr(string, '?'))
    {
        F_("The string \"%s\" contains wildcards line '*' and '?'!\n",
           string);
        return (FALSE);
    }

    /* if existing service is a pattern */
    if (strchr(pattern, '*') || strchr(pattern, '?'))
    {
        if (fnmatch(pattern, string, FNM_CASEFOLD) == 0)

        {
            /* found */
            return (TRUE);
        }
    }

    /* create service name mask for pattern matching */
    service_mask = i_calloc(strlen(string) + 3, sizeof(char));
    assert(service_mask);

    strcpy(service_mask, "*/");
    strcat(service_mask, string);

    if (fnmatch(service_mask, pattern, FNM_CASEFOLD) == 0)
    {
        free(service_mask);
        return (TRUE);
    }

    /* did not find any */
    free(service_mask);
    return (FALSE);
}
