#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>

#include <stdlib.h>

#include "test_initng.h"

#include "../src/initng_load_module.h"



static void test_open_module(void)
{
    m_h *m;

    NEW_TEST();

    INIT_LIST_HEAD(&g.module_db.list);

    INFO("set g.Argv, its needed to when open_module check to blacklist ...\n");
    g.Argv = (char **) calloc(10, sizeof(char *));
    g.Argv[0] = NULL;
    g.Argv[1] = NULL;

    TEST("test a non-existant plugin ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_nonexistant.so",
                                "nonexistant");
    CU_ASSERT(m == NULL);

    TEST("test the simplest plugin ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_empty.so",
                                "plug_empty");
    CU_ASSERT(m != NULL);

    if (m != NULL)
    {
        CU_ASSERT_STRING_EQUAL(m->module_name, "plug_empty");
        CU_ASSERT(m->module_dlhandle != NULL);
        CU_ASSERT(m->module_init != NULL);
        CU_ASSERT(m->module_unload != NULL);
        CU_ASSERT(m->module_needs == NULL);
        close_module(m);
    }

    TEST("missing module_init() ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_broken_0.so",
                                "plug_broken_0");
    CU_ASSERT(m == NULL);

    TEST("missing module_unload() ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_broken_1.so",
                                "plug_broken_1");
    CU_ASSERT(m == NULL);

    TEST("module_needs specified, but empty ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_needs.so",
                                "plug_needs");
    CU_ASSERT(m != NULL);
    if (m != NULL)
    {
        CU_ASSERT_STRING_EQUAL(m->module_name, "plug_needs");
        CU_ASSERT(m->module_dlhandle != NULL);
        CU_ASSERT(m->module_init != NULL);
        CU_ASSERT(m->module_unload != NULL);
        CU_ASSERT(m->module_needs != NULL);
        if (m->module_needs)
        {
            CU_ASSERT(m->module_needs[0] == NULL);
        }
        close_module(m);
    }

    TEST("module_needs specified, non-empty ...\n");
    m = initng_load_module_open(INITNG_PLUGIN_DIR "/libplug_needs_empty.so",
                                "plug_needs_empty");
    CU_ASSERT(m != NULL);
    if (m != NULL)
    {
        CU_ASSERT_STRING_EQUAL(m->module_name, "plug_needs_empty");
        CU_ASSERT(m->module_dlhandle != NULL);
        CU_ASSERT(m->module_init != NULL);
        CU_ASSERT(m->module_unload != NULL);
        CU_ASSERT(m->module_needs != NULL);
        if (m->module_needs)
        {
            CU_ASSERT_STRING_EQUAL(m->module_needs[0], "plug_empty");
            CU_ASSERT(m->module_needs[1] == NULL);
        }
        close_module(m);
    }

    free(g.Argv);
}
static void print_modules(void)
{
    m_h *current = NULL;

    printf(" -- Current modules: \n");
    while_module_db(current)
    {
        printf(" ** %s\n", current->module_name);
    }
    printf(" -- end\n");
}

static void test_initng_load_module(void)
{
    m_h *m;
    m_h *current;

    NEW_TEST();

    INIT_LIST_HEAD(&g.module_db.list);

    INFO("set g.Argv, its needed to when open_module check to blacklist ...\n");
    g.Argv = (char **) calloc(10, sizeof(char *));
    g.Argv[0] = NULL;
    g.Argv[1] = NULL;


    TEST("verify we start empty ...\n");
    CU_ASSERT(list_empty(&g.module_db.list)) print_modules();

    TEST("test the simplest plugin\n");
    m = initng_load_module("plug_empty");
    print_modules();
    CU_ASSERT(m != NULL);

    /* search the db, with only one entry, it should not be hard to find */
    current = NULL;
    while_module_db(current)
    {
        CU_ASSERT(current == m) if (current->module_name)
        {
            CU_ASSERT_STRING_EQUAL(current->module_name, "plug_empty");
        }
    }


    TEST("verify we can't load it twice ...\n");
    m = initng_load_module("plug_empty");
    print_modules();
    CU_ASSERT(m == NULL);

    TEST("try loading a bogus module ...\n");
    m = initng_load_module("plug_nonexistant");
    print_modules();
    CU_ASSERT(m == NULL);

    TEST("verify that we can load a module that has all needs ...\n");
    m = initng_load_module("plug_needs_empty");
    print_modules();
    CU_ASSERT(m != NULL);

    TEST("unload module plug_needs_empty ...\n");
    initng_unload_module_named("plug_needs_empty");
    print_modules();
    TEST("unload all modules...\n");
    initng_unload_module_unload_all();
    print_modules();

    TEST("verify that we can't load a module that does *NOT* have all it needs ...\n");
    m = initng_load_module("plug_needs_empty");
    print_modules();
    CU_ASSERT(m == NULL);

    free(g.Argv);
}

static void test_initng_unload_module_named(void)
{
    NEW_TEST();

    INIT_LIST_HEAD(&g.module_db.list);

    INFO("set g.Argv, its needed to when open_module check to blacklist ...\n");
    g.Argv = (char **) calloc(10, sizeof(char *));
    g.Argv[0] = NULL;
    g.Argv[1] = NULL;


    TEST("verify we start empty ...\n");
    CU_ASSERT(list_empty(&g.module_db.list))
        TEST("try unloaded a module that is not loaded ...\n");
    CU_ASSERT(initng_unload_module_named("not_loaded") == FALSE);

    TEST("try a load and unload ...\n");
    CU_ASSERT(initng_load_module("plug_empty") != NULL);
    CU_ASSERT(initng_unload_module_named("plug_empty") == TRUE);

    TEST("verify we can't unload again ...\n");
    CU_ASSERT(initng_unload_module_named("plug_empty") == FALSE);

    TEST("try to unload a needed module ...\n");
    CU_ASSERT(initng_load_module("plug_empty") != NULL);
    CU_ASSERT(initng_load_module("plug_needs_empty") != NULL);
    CU_ASSERT(initng_unload_module_named("plug_empty") == FALSE);

    TEST("verify we can unload in the proper order ...\n");
    CU_ASSERT(initng_unload_module_named("plug_needs_empty") == TRUE);
    CU_ASSERT(initng_unload_module_named("plug_empty") == TRUE);

    free(g.Argv);
}

static void test_initng_load_all_modules(void)
{
    m_h *m;
    int found = 0;

    NEW_TEST();


    /* this sets g.Argv */
    initng_fake_new();


    TEST(" make sure the db is empty ");
    m = NULL;
    while_module_db(m)
    {
        CU_ASSERT(m == NULL);
    }


    TEST(" load the modules ");
    CU_ASSERT(initng_load_module_load_all());

    /* 
     * our list should be one of:
     *  needs_empty -> needs -> empty
     *  needs_empty -> empty -> needs
     *  needs -> needs_empty -> empty
     */

    TEST("look for the 2 modules ");
    printf("Look for modules, that we want to be loaded.");
    m = NULL;
    while_module_db(m)
    {
        if (strcmp(m->module_name, "plug_needs") == 0)
        {
            printf("Found %s!!!\n", m->module_name);
            sleep(1);
            found++;
        }
    }

    TEST("look if they both was found");
    CU_ASSERT(found == 1);

    INFO(" unload the modules ");
    initng_unload_module_unload_all();

    TEST(" make sure the db is empty ");
    m = NULL;
    while_module_db(m)
    {
        CU_ASSERT(m == NULL);
    }

    free(g.Argv);

}

void test_initng_load_module_setup(void)
{
    CU_pSuite suite;

    suite = CU_add_suite("initng_load_module", empty_init, empty_cleanup);
    CU_add_test(suite, "open_module", test_open_module);
    CU_add_test(suite, "initng_load_module", test_initng_load_module);
    CU_add_test(suite, "initng_unload_module_named",
                test_initng_unload_module_named);
    CU_add_test(suite, "initng_load_all_modules",
                test_initng_load_all_modules);
}
